/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */

#include <config.h>

#ifdef HAVE_DBUS
#include <dbus/dbus-glib.h>
#endif

#include <gtk/gtk.h>

#include <main.h>
#include "error.h"

#include <bmp_ui.h>
#include <bmp_play.h>
#include <bmp_playlist.h>
#include <bmp_resource_manager.h>
#include <bmp_system_control.h>
#include <logger.h>
#include <g_object_get.h>
#include <ui_util.h>

#include <bmp_marshalers.h>

G_DEFINE_TYPE (BmpSystemControl, bmp_system_control, G_TYPE_OBJECT);

static gboolean proceed = TRUE;

typedef enum {
    BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_TRACKLIST,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME,

    BMP_SYSTEM_CONTROL_SIGNAL_LIST_DELETED,
    BMP_SYSTEM_CONTROL_SIGNAL_LIST_ADDED,
    BMP_SYSTEM_CONTROL_SIGNAL_LIST_RENAMED,

    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_LIST_SORTED,
    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED,
    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED,

    BMP_SYSTEM_CONTROL_N_SIGNALS

} BmpSystemControlSignals;

static guint signals[BMP_SYSTEM_CONTROL_N_SIGNALS] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

typedef enum {
    BMP_SYSTEM_CONTROL_PROPERTY_VOLUME = 1,
    BMP_SYSTEM_CONTROL_PROPERTY_BITRATE,
    BMP_SYSTEM_CONTROL_PROPERTY_SAMPLERATE,
    BMP_SYSTEM_CONTROL_PROPERTY_STREAM_POS,

    BMP_SYSTEM_CONTROL_PROPERTY_MOVING,
    BMP_SYSTEM_CONTROL_PROPERTY_REPEAT,
    BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE,

    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TRACKLIST,
    BMP_SYSTEM_CONTROL_PROPERTY_ACTIVE_TRACKLIST,

    BMP_SYSTEM_CONTROL_N_PROPERTIES
} BmpSystemControlProperties;

struct _BmpSystemControlPrivate {
    gboolean dispose_has_run;

    GAsyncQueue *message_queue;

    /* Properties */

    gint bitrate;
    gint current_list;
    gint active_list;
    gint current_track;
    gint playlist;
    gint samplerate;
    gint stream_pos;
    gint volume;

    gboolean moving;
    gboolean repeat;
    gboolean shuffle;
};

typedef struct {
    gint list;
    gint track;
} bmp_message_list_track_t;

typedef struct {
    BmpEqEnum eq;
    gint value;
} bmp_message_eq_t;

static gint eq_values[BMP_N_EQ_PARAMS];

static gchar *eq_configuration_keynames[] = {
    "eq-30",
    "eq-60",
    "eq-125",
    "eq-250",
    "eq-500",
    "eq-1000",
    "eq-2000",
    "eq-4000",
    "eq-8000",
    "eq-16000"
};

static gint n_eq_configuration_keynames =
G_N_ELEMENTS (eq_configuration_keynames);

#ifdef HAVE_DBUS

static gboolean _no_remote = FALSE;

/* DBus methods */
gboolean
bmp_system_control_get_playlists (BmpSystemControl * self,
				  GValueArray ** array,
				  GError ** error);
gboolean
bmp_system_control_get_playlist (BmpSystemControl * self,
				 gint tracklist,
				 GValueArray ** array,
				 GError ** error);
gboolean
bmp_system_control_get_playlist_name (BmpSystemControl * self,
				 gint tracklist,
				 gchar **name,
				 GError ** error);

gboolean
bmp_system_control_get_current_title (BmpSystemControl * self,
				      gchar ** title,
				      GError ** error);
gboolean
bmp_system_control_get_current_uri (BmpSystemControl * self,
				    gchar ** uri, GError ** error);
gboolean
bmp_system_control_get_current_track (BmpSystemControl * self,
				      gint * idx, GError ** error);

gboolean
bmp_system_control_get_metadata_for_uri (BmpSystemControl * self,
					 gchar * uri,
					 GHashTable ** metadata,
					 GError ** error);

gboolean
bmp_system_control_get_metadata_for_list_item (BmpSystemControl * self,
					 gint list,
					 gint item, 
					 GHashTable ** metadata,
					 GError ** error);

gboolean
bmp_system_control_get_current_tracklist (BmpSystemControl * self,
					  gint * tracklist,
					  GError ** error);

gboolean
bmp_system_control_send_status (BmpSystemControl * self,
				GError ** error);

gboolean
bmp_system_control_volume_get (BmpSystemControl * self,
			       gint * volume, GError ** error);
gboolean
bmp_system_control_eq_get (BmpSystemControl * self, gint eq,
			   gint * value, GError ** error);

#include <bmp_system_control_glue.h>

#endif

gboolean
bmp_system_control_send_status (BmpSystemControl * self, GError ** error)
{
    g_signal_emit (G_OBJECT (self),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS],
		   0, g_object_get_int (bmp_play, "playstatus"));

    return TRUE;
}

static void
ht_copy (gpointer key, gpointer value, GHashTable * _copy)
{
    g_hash_table_insert (_copy, g_strdup (key), g_strdup (value));
}

gboolean
bmp_system_control_get_metadata_for_uri (BmpSystemControl * self,
					 gchar * uri,
					 GHashTable ** metadata,
					 GError ** error)
{
    GHashTable *_metadata, *_metadata_orig;

    _metadata = g_hash_table_new (g_str_hash, g_str_equal);
    _metadata_orig =
      bmp_metadata_cache_get_metadata_for_uri (bmp_metadata_cache, uri);

    g_hash_table_foreach (_metadata_orig, (GHFunc) ht_copy, _metadata);

    *metadata = _metadata;

    return TRUE;
}

gboolean
bmp_system_control_get_metadata_for_list_item (BmpSystemControl * self,
					 gint list,
				         gint item, 
					 GHashTable ** metadata,
					 GError ** error)
{
    GHashTable *_metadata, *_metadata_orig;
    BmpListType *list_type;
    BmpFileItem *file_item; 

    list_type = 
	bmp_playlist_tracklist_get_i (bmp_playlist, list);
    file_item = 
	g_list_nth (list_type->items, item)->data;

    _metadata = g_hash_table_new (g_str_hash, g_str_equal);
    _metadata_orig =
      bmp_metadata_cache_get_metadata_for_uri (bmp_metadata_cache, file_item->uri);

    g_hash_table_foreach (_metadata_orig, (GHFunc) ht_copy, _metadata);

    *metadata = _metadata;

    return TRUE;
}


gboolean
bmp_system_control_add_uri_list (BmpSystemControl * self,
				 gchar ** uri_list,
				 gint tracklist,
				 gint position,
				 gint * position_out,
				 GError ** error)
{
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL (self), TRUE);

    if (tracklist == -1)
	{
	    tracklist = g_object_get_int (self, "current_tracklist");
	}

    if (position == -1)
	{
	    *position_out = bmp_playlist_insert_tracklist_items (bmp_playlist,
								 tracklist,
								 (const gchar
								  **) uri_list);
	}
    else
	{
	    *position_out =
	      bmp_playlist_insert_tracklist_items_position (bmp_playlist,
							    tracklist,
							    (const gchar **)
							    uri_list,
							    position);
	}

    return TRUE;
}

gboolean
bmp_system_control_current_tracklist_get (BmpSystemControl * self,
					  gint * tracklist, GError ** error)
{
    *tracklist = self->private->current_list;

    return TRUE;
}

gboolean
bmp_system_control_eq_get (BmpSystemControl * self,
			   gint eq, gint * value, GError ** error)
{
    *value = eq_values[eq];

    return TRUE;
}

gboolean
bmp_system_control_volume_get (BmpSystemControl * self,
			       gint * volume, GError ** error)
{
    *volume = self->private->volume;

    return TRUE;
}

gboolean
bmp_system_control_get_current_title (BmpSystemControl * self,
				      gchar ** title, GError ** error)
{
    *title = g_strdup (bmp_playlist_get_current_title (bmp_playlist));

    return TRUE;
}

gboolean
bmp_system_control_get_current_tracklist (BmpSystemControl * self,
					  gint * tracklist, GError ** error)
{
    *tracklist = g_object_get_int (self, "current_tracklist");

    return TRUE;
}

gboolean
bmp_system_control_get_current_uri (BmpSystemControl * self,
				    gchar ** uri, GError ** error)
{
    *uri = g_strdup (bmp_playlist_get_current_uri (bmp_playlist));

    return TRUE;
}

gboolean
bmp_system_control_get_current_track (BmpSystemControl * self,
				      gint * idx, GError ** error)
{
    *idx = bmp_playlist_get_current_track_idx (bmp_playlist);

    return TRUE;
}

gboolean
bmp_system_control_get_playlist_name (BmpSystemControl * self,
				 gint tracklist,
				 gchar **name,
				 GError ** error)
{
    *name = g_strdup(bmp_playlist_get_nth_tracklist_name (bmp_playlist, tracklist));

    return TRUE;
}
gboolean
bmp_system_control_get_playlist (BmpSystemControl * self,
				 gint tracklist,
				 GValueArray ** array,
				 GError ** error)
{
    BmpListType *list;
    BmpFileItem *file_item;
    GList *items;
    GValue value = { 0, };
    gint length;

    list = bmp_playlist_tracklist_get_i (bmp_playlist, tracklist);
    items = list->items;
    length = g_list_length (items);
    *array = g_value_array_new (length);

    g_value_init (&value, G_TYPE_STRING);

    while (items)
	{
	    file_item = (BmpFileItem *) items->data;
	    g_value_set_string (&value, file_item->uri);
	    g_value_array_append (*array, &value);

	    items = g_list_next (items);
	}

    return TRUE;
}

gboolean
bmp_system_control_get_playlists (BmpSystemControl * self,
				  GValueArray ** array,
				  GError ** error)
{
    BmpListType *list;
    BmpPlaylistItem *playlist_item;
    GList *items;
    GValue value = { 0, };
    gint length;

    list = bmp_playlist_get_playlists (bmp_playlist);
    items = list->items;
    length = g_list_length (items);

    if (!length)
	{
	    *array = NULL;
	    *error =
	      g_error_new_literal (g_quark_from_string ("BmpSystemControl"),
				   1, "No Lists");
	    return FALSE;
	}

    *array = g_value_array_new (length);

    g_value_init (&value, G_TYPE_STRING);

    while (items)
	{
	    playlist_item = (BmpPlaylistItem *) items->data;
	    g_value_set_string (&value, playlist_item->listname);
	    g_value_array_append (*array, &value);

	    items = g_list_next (items);
	}

    return TRUE;
}

static void
free_message_list_track (gpointer msgpointer)
{
    bmp_message_list_track_t *message =
      (bmp_message_list_track_t *) msgpointer;

    g_free (message);
}

/* Our stuff */
static gboolean
play_list_track (BmpSystemControl * self, gint _idx_list, gint _idx)
{
    BmpListType *tracklist;
    BmpListType *playlists;
    BmpFileItem *file_item;
    gchar *mrl;

    bmp_playlist_clear_all_current (bmp_playlist);

    tracklist = bmp_playlist_tracklist_get_i (bmp_playlist, _idx_list);
    if (!tracklist)
	return FALSE;
    if (!tracklist->items)
	return FALSE;

    playlists = bmp_playlist_get_playlists (bmp_playlist);
    if (!playlists)
	return FALSE;

    file_item = LIST_GET_NTH_ITEM (tracklist, _idx);
    if (!file_item)
	return FALSE;

    g_object_set (self, "active_tracklist", _idx_list, NULL);
    self->private->current_track = _idx;

    mrl = file_item->uri;

    tracklist->current = _idx;
    playlists->current = _idx_list;

    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_STOPPED,
		  NULL);
    g_object_set (G_OBJECT (bmp_play), "mrl", mrl, NULL);
    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_PLAYING,
		  NULL);

    g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE], 0);

    return TRUE;
}

static void
play_current (BmpSystemControl * self)
{
    BmpListType *tracklist;

    tracklist =
      bmp_playlist_tracklist_get_i (bmp_playlist,
				    self->private->current_list);

    if (tracklist->current == -1)
	tracklist->current = 0;

    if (tracklist)
	play_list_track (self, self->private->current_list,
			 tracklist->current);
}

static void
play_next (BmpSystemControl * self)
{
    BmpListType *tracklist;
    BmpFileItem *file_item;
    gboolean repeat;
    gboolean shuffle;
    gchar *mrl;
    gint _idx;

    repeat = g_object_get_boolean (self, "repeat");
    shuffle = g_object_get_boolean (self, "shuffle");

    if (!shuffle)
	{
	    tracklist = bmp_playlist_tracklist_get_i (bmp_playlist,
						      self->private->active_list);

	    if (!tracklist)
		{
		    g_object_set (G_OBJECT (bmp_play), "playstatus",
				  PLAYSTATUS_STOPPED, NULL);
		    return;
		}

	    if (!tracklist->items)
		{
		    tracklist->current = -1;
		    g_object_set (G_OBJECT (bmp_play), "playstatus",
				  PLAYSTATUS_STOPPED, NULL);
		    return;
		}

	    tracklist->current++;

	    if (tracklist->current > (g_list_length (tracklist->items) - 1))
		{
		    if (!repeat)
			{
			    tracklist->current = -1;
			    g_object_set (G_OBJECT (bmp_play), "playstatus",
					  PLAYSTATUS_STOPPED, NULL);
			    return;
			}

		    tracklist->current = 0;
		}

	    _idx = tracklist->current;

	}
    else
	{
	    tracklist = bmp_playlist_tracklist_get_i (bmp_playlist,
						      self->private->active_list);

	    if (!tracklist)
		{
		    g_object_set (G_OBJECT (bmp_play), "playstatus",
				  PLAYSTATUS_STOPPED, NULL);
		    return;
		}

	    if (!tracklist->items)
		{
		    tracklist->current = -1;
		    g_object_set (G_OBJECT (bmp_play), "playstatus",
				  PLAYSTATUS_STOPPED, NULL);
		    return;
		}

	    tracklist->current_linear++;

	    if (tracklist->current_linear >
		(g_list_length (tracklist->items_shuffle) - 1))
		{
		    if (!repeat)
			{
			    tracklist->current_linear = -1;
			    tracklist->current = tracklist->current_linear;
			    g_object_set (G_OBJECT (bmp_play), "playstatus",
					  PLAYSTATUS_STOPPED, NULL);
			    return;
			}

		    g_object_set (G_OBJECT (self), "shuffle", TRUE, NULL);
		}

	    g_assert (tracklist->items_shuffle != NULL);

	    _idx = GPOINTER_TO_INT (g_list_nth (tracklist->items_shuffle,
						tracklist->current_linear)->data);
	    tracklist->current = _idx;
	}

    self->private->current_track = _idx;

    file_item = g_list_nth (tracklist->items, _idx)->data;

    if (!file_item)
	{
	    tracklist->current = -1;
	    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_STOPPED,
			  NULL);
	    return;
	}

    mrl = file_item->uri;

    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_STOPPED,
		  NULL);
    g_object_set (G_OBJECT (bmp_play), "mrl", mrl, NULL);
    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_PLAYING,
		  NULL);

    g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE], 0);
}

static void
play_prev (BmpSystemControl * self)
{
    BmpListType *tracklist;
    BmpFileItem *file_item;
    gboolean repeat;
    gboolean shuffle;
    gchar *mrl;
    gint _idx;

    repeat = g_object_get_boolean (self, "repeat");
    shuffle = g_object_get_boolean (self, "shuffle");

    if (!shuffle)
	{
	    tracklist =
	      bmp_playlist_tracklist_get_i (bmp_playlist,
					    self->private->active_list);

	    if (!tracklist->items)
		{
		    tracklist->current = -1;
		    g_object_set (G_OBJECT (bmp_play), "playstatus",
				  PLAYSTATUS_STOPPED, NULL);
		    return;
		}

	    tracklist->current--;

	    if (tracklist->current < 0)
		{
		    if (!repeat)
			{
			    tracklist->current = -1;
			    g_object_set (G_OBJECT (bmp_play), "playstatus",
					  PLAYSTATUS_STOPPED, NULL);
			    return;
			}

		    tracklist->current = g_list_length (tracklist->items) - 1;
		}

	    _idx = tracklist->current;

	}
    else
	{
	    tracklist =
	      bmp_playlist_tracklist_get_i (bmp_playlist,
					    self->private->active_list);

	    if (!tracklist->items)
		{
		    tracklist->current = -1;
		    g_object_set (G_OBJECT (bmp_play), "playstatus",
				  PLAYSTATUS_STOPPED, NULL);
		    return;
		}

	    tracklist->current_linear--;

	    if (tracklist->current_linear < 0)
		{
		    if (!repeat)
			{
			    tracklist->current_linear = -1;
			    tracklist->current = tracklist->current_linear;
			    g_object_set (G_OBJECT (bmp_play), "playstatus",
					  PLAYSTATUS_STOPPED, NULL);
			    return;
			}

		    tracklist->current_linear =
		      g_list_length (tracklist->items_shuffle) - 1;
		}

	    _idx =
	      GPOINTER_TO_INT (g_list_nth
			       (tracklist->items_shuffle,
				tracklist->current_linear)->data);
	    tracklist->current = _idx;
	}

    self->private->current_track = _idx;

    file_item = g_list_nth (tracklist->items, _idx)->data;

    if (!file_item)
	{
	    tracklist->current = -1;
	    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_STOPPED,
			  NULL);
	    return;
	}

    mrl = file_item->uri;

    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_STOPPED,
		  NULL);
    g_object_set (G_OBJECT (bmp_play), "mrl", mrl, NULL);
    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_PLAYING,
		  NULL);

    g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE], 0);

    return;
}

static void
play_pause (BmpSystemControl * self)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_PAUSED, NULL);
}

static void
play_stop (BmpSystemControl * self)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    g_object_set (G_OBJECT (bmp_play), "playstatus", PLAYSTATUS_STOPPED,
		  NULL);
    g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE], 0);
}

static void
play_seek (BmpSystemControl * self, gint position)
{
    gint length, position_new;
    gdouble percent;

    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    length = bmp_play_get_length (bmp_play);
    percent = length / 100.;

    position_new = position * percent;

    bmp_play_seek_stream (bmp_play, position_new * 1000);
}

static void
play_volume (BmpSystemControl * self, gint volume)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    bmp_play_set_volume (bmp_play, volume);
    g_object_set (self, "volume", volume, NULL);
}

static void
play_eq (BmpSystemControl * self, BmpEqEnum eq, gint value)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

#ifdef HAVE_XINE
    bmp_play_set_eq (bmp_play, eq, value);
#endif
}

static void
save_configuration (BmpSystemControl * self)
{
    BmpWindowPlaylist *window_playlist;
    BmpWindowMain *window_main;
    gint x, y, w, h;

    window_main = RM_GET_ITEM ("windows", "window-main", NOLOCK);
    window_playlist = RM_GET_ITEM ("windows", "window-playlist", NOLOCK);

    /* SAVE CONFIGURATION */

    bmp_configuration_key_set_int (bmp_configuration, "bmp", "volume",
				   self->private->volume);

    if (bmp_ui)
	{
	    /* Main window */
	    gtk_window_get_position (GTK_WINDOW (window_main->window), &x, &y);
	    bmp_configuration_key_set_int (bmp_configuration, "main-window",
					   "pos_x", x);
	    bmp_configuration_key_set_int (bmp_configuration, "main-window",
					   "pos_y", y);

	    /* Playlist window */
	    gtk_window_get_position (GTK_WINDOW (window_playlist->window), &x,
				     &y);
	    bmp_configuration_key_set_int (bmp_configuration, "playlist-window",
					   "pos_x", x);
	    bmp_configuration_key_set_int (bmp_configuration, "playlist-window",
					   "pos_y", y);

	    gtk_window_get_size (GTK_WINDOW (window_playlist->window), &w, &h);
	    bmp_configuration_key_set_int (bmp_configuration, "playlist-window",
					   "width", w);
	    bmp_configuration_key_set_int (bmp_configuration, "playlist-window",
					   "height", h);
	}
}

static gboolean
process_messages (BmpSystemControl * self)
{
    bmp_message_t *message;
    gint queue_length;

    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL (self), FALSE);

    message = g_async_queue_try_pop (self->private->message_queue);

    if (!message)
	return FALSE;

    switch (message->message_type)
	{
	case BMP_SYSTEM_MESSAGE_PAUSE:
		{
		    play_pause (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_STOP:
		{
		    play_stop (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_NEXT:
		{
		    play_next (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_CURRENT:
		{
		    play_current (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_PREV:
		{
		    play_prev (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_PLAY_LIST_TRACK:
		{
		    bmp_message_list_track_t *message_data = message->user_data;

		    play_list_track (self, message_data->list, message_data->track);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_PLAY_TRACK:
		{
		    gint _idx_list;

		    _idx_list = g_object_get_int (self, "current_tracklist");

		    play_list_track (self, _idx_list,
				     GPOINTER_TO_INT (message->user_data));
		}
	    break;

	case BMP_SYSTEM_MESSAGE_PLAY_LIST:
		{
		    play_list_track (self, GPOINTER_TO_INT (message->user_data), 0);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SEEK:
		{
		    play_seek (self, GPOINTER_TO_INT (message->user_data));
		}
	    break;

	case BMP_SYSTEM_MESSAGE_VOLUME:
		{
		    play_volume (self, GPOINTER_TO_INT (message->user_data));
		}
	    break;

	case BMP_SYSTEM_MESSAGE_EQ:
		{
		    bmp_message_eq_t *message_eq;

		    message_eq = message->user_data;

		    play_eq (self, message_eq->eq, message_eq->value);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_BITRATE:
		{
		    g_object_set (self, "bitrate",
				  GPOINTER_TO_INT (message->user_data), NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SAMPLERATE:
		{
		    g_object_set (self, "samplerate",
				  GPOINTER_TO_INT (message->user_data), NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SAVE_STATE:
		{
		    save_configuration (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SHOW_PL_WINDOW:
		{
		    gtk_toggle_action_set_active (GTK_TOGGLE_ACTION
						  (gtk_action_group_get_action
						   (bmp_toggle_actions,
						    BMP_TOGGLE_ACTION_PL)),
						  GPOINTER_TO_INT (message->
								   user_data));
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SET_CURRENT_TRACKLIST:
		{
		    g_object_set (self, "current_tracklist",
				  GPOINTER_TO_INT (message->user_data), NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_REPEAT:
		{
		    gboolean active;

		    active = GPOINTER_TO_INT (message->user_data);

		    g_object_set (G_OBJECT (self), "repeat", active, NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SHUFFLE:
		{
		    gboolean active;

		    active = GPOINTER_TO_INT (message->user_data);

		    g_object_set (G_OBJECT (self), "shuffle", active, NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_QUIT:
		{
		    g_message (" ");
		    g_message ("SHUTTING DOWN");
		    g_message ("[%s]", get_timestamp_str ());
		    g_message (" ");
		    save_configuration (self);

		    if (bmp_ui)
			{
			    g_object_unref (bmp_ui);
			    bmp_ui = NULL;
			}

		    gtk_main_quit ();
		}
	    break;

	case BMP_SYSTEM_MESSAGE_UI_START:
		{
		    bmp_ui = bmp_ui_new ();
		    if (bmp_ui_finalize_init (bmp_ui) != BMP_ERROR_SUCCESS)
			{
			    g_object_unref (bmp_ui);
			    bmp_ui = NULL;
			    gtk_main_quit ();
			}
		}
	    break;

	case BMP_SYSTEM_MESSAGE_UI_STOP:
		{
		    if (bmp_ui)
			g_object_unref (bmp_ui);
		}
	    break;
	}

    queue_length = g_async_queue_length (self->private->message_queue);

    if (queue_length == 0)
	return FALSE;

    return TRUE;
}

/* Public API */
void
bmp_system_message_send (BmpSystemControl * self,
			 BmpSystemMessage message_type,
			 gpointer user_data, GDestroyNotify user_data_destroy)
{
    bmp_message_t *message;

    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    message = g_new0 (bmp_message_t, 1);

    message->message_type = message_type;
    message->user_data = user_data;
    message->user_data_destroy = user_data_destroy;

    g_async_queue_push (self->private->message_queue, message);

    g_idle_add ((GSourceFunc) process_messages, self);
}

void
bmp_system_control_current_tracklist_set (BmpSystemControl * self, gint _idx)
{
    g_object_set (self, "current_tracklist", _idx, NULL);
}

void
bmp_system_control_show_pl_window (BmpSystemControl * self, gboolean show)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_SHOW_PL_WINDOW,
			     GINT_TO_POINTER (show), NULL);
}

void
bmp_system_control_repeat (BmpSystemControl * self, gboolean repeat)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_REPEAT,
			     GINT_TO_POINTER (repeat), NULL);
}

void
bmp_system_control_shuffle (BmpSystemControl * self, gboolean shuffle)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_SHUFFLE,
			     GINT_TO_POINTER (shuffle), NULL);
}

void
bmp_system_control_move_start (BmpSystemControl * self)
{
    g_object_set (self, "moving", TRUE, NULL);
}

void
bmp_system_control_move_end (BmpSystemControl * self)
{
    g_object_set (self, "moving", FALSE, NULL);
}

void
bmp_system_control_ui_start (BmpSystemControl * self)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_UI_START, NULL, NULL);
}

void
bmp_system_control_ui_stop (BmpSystemControl * self)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_UI_STOP, NULL, NULL);
}

void
bmp_system_control_quit (BmpSystemControl * self)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_QUIT, NULL, NULL);
}

void
bmp_system_control_volume_set (BmpSystemControl * self, gint volume)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_VOLUME,
			     GINT_TO_POINTER (volume), NULL);
}

void
bmp_system_control_eq_set (BmpSystemControl * self, BmpEqEnum eq, gint value)
{
    bmp_message_eq_t *message;

    message = g_new0 (bmp_message_eq_t, 1);

    message->eq = eq;
    message->value = value;

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_EQ, message, g_free);
}

void
bmp_system_control_save_state (BmpSystemControl * self)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_SAVE_STATE, NULL, NULL);
}

void
bmp_system_control_play_track (BmpSystemControl * self, gint _idx)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_PLAY_TRACK,
			     GINT_TO_POINTER (_idx), NULL);
}

void
bmp_system_control_play_list (BmpSystemControl * self, gint _idx)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_PLAY_LIST,
			     GINT_TO_POINTER (_idx), NULL);
}

void
bmp_system_control_play_list_track (BmpSystemControl * self, gint _idx_list,
				    gint _idx)
{
    bmp_message_list_track_t *message_data;

    message_data = g_new0 (bmp_message_list_track_t, 1);

    message_data->list = _idx_list;
    message_data->track = _idx;

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_PLAY_LIST_TRACK,
			     message_data, free_message_list_track);
}

void
bmp_system_control_samplerate (BmpSystemControl * self, gint samplerate)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_SAMPLERATE,
			     GINT_TO_POINTER (samplerate), NULL);
}

void
bmp_system_control_bitrate (BmpSystemControl * self, gint bitrate)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_BITRATE,
			     GINT_TO_POINTER (bitrate), NULL);
}

void
bmp_system_control_seek (BmpSystemControl * self, gint position)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_SEEK,
			     GINT_TO_POINTER (position), NULL);
}

void
bmp_system_control_play_pause (BmpSystemControl * self)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_PAUSE, NULL, NULL);
}

gboolean
bmp_system_control_play_stop (BmpSystemControl * self, GError * error)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_STOP, NULL, NULL);

    return TRUE;
}

gboolean
bmp_system_control_play_prev (BmpSystemControl * self, GError * error)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_STOP, NULL, NULL);
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_PREV, NULL, NULL);

    return TRUE;
}

gboolean
bmp_system_control_play_next (BmpSystemControl * self, GError * error)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_STOP, NULL, NULL);
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);

    return TRUE;
}

void
bmp_system_control_play_current (BmpSystemControl * self)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_CURRENT, NULL, NULL);
}

static void
on_play_eos (BmpPlay * play, BmpSystemControl * control)
{
    bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
}

static void
on_play_playstatus (BmpPlay * play, GParamSpec * spec,
		    BmpSystemControl * control)
{
    g_signal_emit (G_OBJECT (control),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS],
		   0, g_object_get_int (bmp_play, "playstatus"));
}


/* GObject stuff */

/* Proxy BmpPlaylist signals */
static void
playlist_callback_list_deleted (BmpPlaylist *playlist, 
				gint list,
				BmpSystemControl *self)
{
    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_DELETED] > 0)
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_DELETED],
                   0, list);
}

static void
playlist_callback_list_added (BmpPlaylist *playlist, 
			      gint list, 
			      BmpSystemControl *self)
{
    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_ADDED] > 0)
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_ADDED],
                   0, list);
}

static void
playlist_callback_list_renamed (BmpPlaylist *playlist, 
				gint list, 
			        BmpSystemControl *self)	
{
    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_RENAMED] > 0)
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_RENAMED],
                   0, list);
}

static void
playlist_callback_tracklist_list_sorted (BmpPlaylist *playlist, 
				         gint list, 
					 BmpSystemControl *self) 
{
    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_LIST_SORTED] > 0)
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_LIST_SORTED],
                   0, list);
}

static void
playlist_callback_tracklist_items_removed (BmpPlaylist *playlist, 
					   gint list, 
					   GArray *removed_rows,  
					   gint n_removed_rows,
					   BmpSystemControl *self)
{
    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED] > 0)
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, list, removed_rows, n_removed_rows);
}

static void
playlist_callback_tracklist_items_added (BmpPlaylist *playlist, 
				         gint list, 
					 GArray *added_rows,  
				         gint n_added_rows,
					 BmpSystemControl *self) 
{
    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED] > 0)
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED],
                   0, list, added_rows, n_added_rows);
}

static void
bmp_system_control_init (BmpSystemControl * self)
{
#ifdef HAVE_DBUS
    DBusGConnection *bus;
    DBusGProxy *bus_proxy;
    gboolean connected = FALSE;
#endif
    GError *error = NULL;
    guint request_name_result;
    gint n;

    self->private = g_new (BmpSystemControlPrivate, 1);

    self->private->dispose_has_run = FALSE;
    self->private->message_queue = g_async_queue_new ();

    self->private->volume =
      bmp_configuration_key_fetch_int (bmp_configuration, "bmp", "volume");
    self->private->bitrate = 0;
    self->private->samplerate = 0;

    self->private->moving = FALSE;
    self->private->repeat = FALSE;
    self->private->shuffle = FALSE;

    self->private->current_track = -1;
    self->private->current_list = -1;

#ifdef HAVE_DBUS
    if (_no_remote)
	{
	    g_message ("%s:\tINITIALIZED", __FILE__);
	    return;
	}

    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);

    if (!bus)
	{
	    g_message ("%s:\tINITIALIZED (DBUS Error: '%s')", __FILE__,
		       error->message);
	    g_error_free (error);
	    return;

	}

    bus_proxy =
      dbus_g_proxy_new_for_name (bus,
				 "org.freedesktop.DBus",
				 "/org/freedesktop/DBus",
				 "org.freedesktop.DBus");

    if (!dbus_g_proxy_call (bus_proxy, "RequestName", &error,
			    G_TYPE_STRING, BMP_DBUS_NAME,
			    G_TYPE_UINT, DBUS_NAME_FLAG_PROHIBIT_REPLACEMENT,
			    G_TYPE_INVALID,
			    G_TYPE_UINT, &request_name_result,
			    G_TYPE_INVALID))
	{
	    g_error ("%s: Failed RequestName request: %s", G_STRFUNC,
		     error->message);
	    g_error_free (error);
	    error = NULL;
	}

    switch (request_name_result)
	{
	case DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER:
		{
		    dbus_g_connection_register_g_object (bus, BMP_DBUS_OBJECT_PATH,
							 G_OBJECT (self));
		    connected = TRUE;
		}
	    break;

	case DBUS_REQUEST_NAME_REPLY_EXISTS:
		{
		    g_message
		      ("%s: BMP session already running, not registering SystemControl object to DBus.",
		       G_STRFUNC);
		    connected = FALSE;
		    proceed = FALSE;
		}
	    break;
	}


#endif

    for (n = 0; n < n_eq_configuration_keynames; n++)
	{
	    LOOKUP_OR_CREATE_DEFAULT_INT ("bmp", eq_configuration_keynames[n],
					  0);
	}

#ifdef HAVE_DBUS
    if (connected)
	{
	    g_message ("%s:\tINITIALIZED  (DBus:%s)", __FILE__,
		       BMP_DBUS_OBJECT_PATH);
	}
    else
	{
	    g_message ("%s:\tINITIALIZED", __FILE__);
	}
#else
    g_message ("%s:\tINITIALIZED", __FILE__);
#endif

}

static void
bmp_system_control_set_property (GObject * object,
				 guint property_id,
				 const GValue * value, GParamSpec * pspec)
{
    BmpSystemControl *self = BMP_SYSTEM_CONTROL (object);

    switch (property_id)
	{
	case BMP_SYSTEM_CONTROL_PROPERTY_STREAM_POS:
		{
		    self->private->stream_pos = g_value_get_int (value);
		    g_signal_emit (self,
				   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS],
				   0, self->private->stream_pos);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_BITRATE:
		{
		    self->private->bitrate = g_value_get_int (value);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_SAMPLERATE:
		{
		    self->private->samplerate = g_value_get_int (value);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_VOLUME:
		{
		    self->private->volume = g_value_get_int (value);
		    g_signal_emit (self,
				   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME],
				   0, self->private->volume);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_MOVING:
		{
		    self->private->moving = g_value_get_boolean (value);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_REPEAT:
		{
		    GtkToggleAction *action;

		    self->private->repeat = g_value_get_boolean (value);
		    action =
		      GTK_TOGGLE_ACTION (gtk_action_group_get_action
					 (bmp_toggle_actions,
					  BMP_TOGGLE_ACTION_REPEAT));

		    if (gtk_toggle_action_get_active (action) !=
			self->private->repeat)
			gtk_toggle_action_set_active (action,
						      self->private->repeat);

		    bmp_configuration_key_set_boolean (bmp_configuration,
						       "bmp", "repeat",
						       self->private->repeat);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE:
		{
		    GtkToggleAction *action;

		    self->private->shuffle = g_value_get_boolean (value);
		    action =
		      GTK_TOGGLE_ACTION (gtk_action_group_get_action
					 (bmp_toggle_actions,
					  BMP_TOGGLE_ACTION_SHUFFLE));

		    if (gtk_toggle_action_get_active (action) !=
			self->private->shuffle)
			gtk_toggle_action_set_active (action,
						      self->private->shuffle);

		    bmp_playlist_tracklist_shuffle (bmp_playlist,
						    g_object_get_int
						    (bmp_system_control,
						     "current_tracklist"));
		    bmp_configuration_key_set_boolean (bmp_configuration,
						       "bmp", "shuffle",
						       self->private->shuffle);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TRACKLIST:
		{
		    gint tracklist;

		    tracklist = g_value_get_int (value);

		    self->private->current_list = tracklist;
		    g_signal_emit (self,
				   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_TRACKLIST],
				   0, tracklist);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_ACTIVE_TRACKLIST:
		{
		    gint tracklist;

		    tracklist = g_value_get_int (value);

		    g_message ("%s: Active tracklist: %d", G_STRFUNC, tracklist);

		    self->private->active_list = tracklist;
		}
	    break;

	default:
	    /* We don't have any other property... */
	    g_assert (FALSE);
	    break;
	}
}

static void
bmp_system_control_get_property (GObject * object,
				 guint property_id,
				 GValue * value, GParamSpec * pspec)
{
    BmpSystemControl *self = BMP_SYSTEM_CONTROL (object);

    switch (property_id)
	{
	case BMP_SYSTEM_CONTROL_PROPERTY_STREAM_POS:
		{
		    g_value_set_int (value, self->private->stream_pos);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_BITRATE:
		{
		    g_value_set_int (value, self->private->bitrate);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_SAMPLERATE:
		{
		    g_value_set_int (value, self->private->samplerate);
		}
	    break;


	case BMP_SYSTEM_CONTROL_PROPERTY_VOLUME:
		{
		    g_value_set_int (value, self->private->volume);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_MOVING:
		{
		    g_value_set_boolean (value, self->private->moving);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_REPEAT:
		{
		    g_value_set_boolean (value, self->private->repeat);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE:
		{
		    g_value_set_boolean (value, self->private->shuffle);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TRACKLIST:
		{
		    g_value_set_int (value, self->private->current_list);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_ACTIVE_TRACKLIST:
		{
		    g_value_set_int (value, self->private->active_list);
		}
	    break;

	default:
	    /* We don't have any other property... */
	    g_assert (FALSE);
	    break;
	}
}

static GObject *
bmp_system_control_constructor (GType type,
				guint n_construct_properties,
				GObjectConstructParam * construct_properties)
{
    GObject *obj;

	{
	    /* Invoke parent constructor. */
	    BmpSystemControlClass *klass;
	    GObjectClass *parent_class;
	    klass =
	      BMP_SYSTEM_CONTROL_CLASS (g_type_class_peek
					(BMP_TYPE_SYSTEM_CONTROL));
	    parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	    obj = parent_class->constructor (type,
					     n_construct_properties,
					     construct_properties);
	}

    /* do stuff. */

    return obj;
}

static void
bmp_system_control_dispose (GObject * obj)
{
    BmpSystemControl *self = (BmpSystemControl *) obj;

    if (self->private->dispose_has_run)
	return;

    self->private->dispose_has_run = TRUE;

    g_message ("%s:\tSTOPPED", __FILE__);
}

static void
bmp_system_control_finalize (GObject * obj)
{
    BmpSystemControl *self = (BmpSystemControl *) obj;

    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */
    g_free (self->private);
}

BmpSystemControl *
bmp_system_control_new (gboolean no_remote)
{
    BmpSystemControl *self;

#ifdef HAVE_DBUS
    if (!no_remote)
	dbus_g_object_type_install_info (BMP_TYPE_SYSTEM_CONTROL,
					 &dbus_glib_bmp_system_control_object_info);
    _no_remote = no_remote;
#endif

    self = g_object_new (bmp_system_control_get_type (), NULL);

    g_signal_connect (G_OBJECT (bmp_play), "eos", G_CALLBACK (on_play_eos),
		      self);
    g_signal_connect (G_OBJECT (bmp_play), "notify::playstatus",
		      G_CALLBACK (on_play_playstatus), self);

    /* Proxy BmpPlaylist signals */
    g_object_connect(G_OBJECT(bmp_playlist),
		"signal::list-deleted",
		G_CALLBACK(playlist_callback_list_deleted),
	 	self,
		"signal::list-added",
		G_CALLBACK(playlist_callback_list_added),
	 	self,
		"signal::list-renamed",
		G_CALLBACK(playlist_callback_list_renamed),
		self,
		"signal::tracklist-list-sorted",
		G_CALLBACK(playlist_callback_tracklist_list_sorted),
	 	self,
		"signal::tracklist-items-removed",
		G_CALLBACK(playlist_callback_tracklist_items_removed),
		self,
		"signal::tracklist-items-added",
		G_CALLBACK(playlist_callback_tracklist_items_added),
		self,
		NULL);

    if (!proceed)
	return NULL;
    else
	return self;
}

static void
bmp_system_control_class_init (BmpSystemControlClass * g_class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
    GParamSpec *pspec;

    gobject_class->set_property = bmp_system_control_set_property;
    gobject_class->get_property = bmp_system_control_get_property;
    gobject_class->dispose = bmp_system_control_dispose;
    gobject_class->finalize = bmp_system_control_finalize;
    gobject_class->constructor = bmp_system_control_constructor;

    /* INT properties */
    pspec =
      g_param_spec_int ("bitrate",
			"bitrate",
			"Current stream's bitrate.",
			0, G_MAXINT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_BITRATE,
				     pspec);

    pspec =
      g_param_spec_int ("samplerate",
			"samplerate",
			"Current stream's samplerate.",
			0, G_MAXINT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_SAMPLERATE,
				     pspec);

    pspec =
      g_param_spec_int ("volume",
			"volume",
			"Current volume.", 0, 100, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_VOLUME,
				     pspec);

    pspec =
      g_param_spec_int ("stream_pos",
			"stream_pos",
			"Current stream position.",
			0, G_MAXINT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_STREAM_POS,
				     pspec);

    pspec =
      g_param_spec_int ("current_tracklist",
			"current_tracklist",
			"Current tracklist",
			-1, G_MAXINT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TRACKLIST,
				     pspec);

    pspec =
      g_param_spec_int ("active_tracklist",
			"active_tracklist",
			"Active (playing) tracklist",
			-1, G_MAXINT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_ACTIVE_TRACKLIST,
				     pspec);

    /* BOOLEAN properties */
    pspec =
      g_param_spec_boolean ("moving",
			    "moving",
			    "Windows moving.", FALSE, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_MOVING,
				     pspec);

    pspec =
      g_param_spec_boolean ("repeat",
			    "repeat",
			    "Player repeat state",
			    FALSE, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_REPEAT,
				     pspec);

    pspec =
      g_param_spec_boolean ("shuffle",
			    "shuffle",
			    "Player shuffle state",
			    FALSE, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE,
				     pspec);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME] =
      g_signal_new ("set-volume",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_volume),
		    NULL, NULL, g_cclosure_marshal_VOID__INT,
		    G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS] =
      g_signal_new ("set-playstatus",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_playstatus),
		    NULL, NULL, g_cclosure_marshal_VOID__INT,
		    G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE] =
      g_signal_new ("track_change",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, track_change),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID,
		    G_TYPE_NONE, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS] =
      g_signal_new ("set-stream-pos",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_stream_pos),
		    NULL, NULL, g_cclosure_marshal_VOID__INT,
		    G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_TRACKLIST] =
      g_signal_new ("set-current-tracklist",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass,
				     set_current_tracklist), NULL, NULL,
		    g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1,
		    G_TYPE_INT);

    /* Proxied signals from BmpPlaylist */
    signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_DELETED] =
        g_signal_new ("list-deleted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, list_deleted),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_ADDED] =
        g_signal_new ("list-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, list_added),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_LIST_RENAMED] =
        g_signal_new ("list-renamed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, list_renamed),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_LIST_SORTED] =
        g_signal_new ("tracklist-list-sorted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_list_sorted),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
        g_signal_new ("tracklist-items-removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_removed),
                      NULL, NULL, g_cclosure_user_marshal_VOID__INT_POINTER_INT, G_TYPE_NONE,
                      3, G_TYPE_INT, DBUS_TYPE_G_INT_ARRAY, G_TYPE_INT);


    signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED] =
        g_signal_new ("tracklist-items-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_added),
                      NULL, NULL, g_cclosure_user_marshal_VOID__INT_POINTER_INT, G_TYPE_NONE,
                      3, G_TYPE_INT, DBUS_TYPE_G_INT_ARRAY, G_TYPE_INT);
}
