#include <config.h>

#include <gdk/gdkprivate.h>
#include <gtk/gtkimage.h>
#include <stdio.h>
#ifndef __USE_BSD
# define __USE_BSD	/* Needed for mkdtemp () */
#endif
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <X11/Xlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#ifdef ENABLE_NLS
# include <libintl.h>
#endif

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

#include <main.h>
#include <util.h>
#include <ui_util.h>
#include <bmp_ui.h>
#include <bmp_configuration.h>
#include <bmp_metadata.h>
#include <uri.h>

#ifdef HAVE_FTS_H
#  include <fts.h>
#endif

static GQuark quark_popup_data;

/*
 * escape_shell_chars ()
 *
 * Escapes characters that are special to the shell inside double quotes.
 */

static GArray*
string_to_garray (const gchar * str);

static gchar*
escape_shell_chars (const gchar * string)
{
    const gchar *special = "$`\"\\";    /* Characters to escape */
    const gchar *in = string;
    gchar *out, *escaped;
    gint num = 0;

    while (*in != '\0')
	if (strchr (special, *in++))
	    num++;

    escaped = g_malloc (strlen (string) + num + 1);

    in = string;
    out = escaped;

    while (*in != '\0')
	{
	    if (strchr (special, *in))
		*out++ = '\\';
	    *out++ = *in++;
	}
    *out = '\0';

    return escaped;
}


/*
 * find <file> in directory <dirname> or subdirectories.  return
 * pointer to complete filename which has to be freed by calling
 * "g_free ()" after use. Returns NULL if file could not be found.
 */

typedef struct {
    const gchar *to_match;
    gchar *match;
    gboolean found;
} FindFileContext;

static gboolean find_file_func (const gchar * path,
				const gchar * basename, gpointer data)
{
    FindFileContext *context = data;

    if (strlen (path) > FILENAME_MAX)
	{
	    g_warning ("Ignoring path: name too long (%s)", path);
	    return TRUE;
	}

    if (g_file_test (path, G_FILE_TEST_IS_REGULAR))
	{
	    if (! g_ascii_strcasecmp (basename, context->to_match))
		{
		    context->match = g_strdup (path);
		    context->found = TRUE;
		    return TRUE;
		}
	}
    else if (g_file_test (path, G_FILE_TEST_IS_DIR))
	{
	    dir_foreach (path, find_file_func, context, NULL);
	    if (context->found)
		return TRUE;
	}

    return FALSE;
}

gchar *find_file_recursively (const gchar * path, const gchar * filename)
{
    FindFileContext context;

    context.to_match = filename;
    context.match = NULL;
    context.found = FALSE;

    dir_foreach (path, find_file_func, &context, NULL);
    return context.match;
}


typedef enum {
    ARCHIVE_UNKNOWN = 0,
    ARCHIVE_DIR,
    ARCHIVE_TAR,
    ARCHIVE_TGZ,
    ARCHIVE_ZIP,
    ARCHIVE_TBZ2
} ArchiveType;

typedef gchar * (*ArchiveExtractFunc) (const gchar *, const gchar *);

typedef struct {
    ArchiveType type;
    const gchar *ext;
} ArchiveExtensionType;

static ArchiveExtensionType archive_extensions[] = {
	{ARCHIVE_TAR, ".tar"},
	{ARCHIVE_ZIP, ".wsz"},
	{ARCHIVE_ZIP, ".zip"},
	{ARCHIVE_TGZ, ".tar.gz"},
	{ARCHIVE_TGZ, ".tgz"},
	{ARCHIVE_TBZ2, ".tar.bz2"},
	{ARCHIVE_TBZ2, ".bz2"},
	{ARCHIVE_UNKNOWN, NULL}
};

static gchar *archive_extract_tar (const gchar * archive,
				   const gchar * dest);
static gchar *archive_extract_zip (const gchar * archive,
				   const gchar * dest);
static gchar *archive_extract_tgz (const gchar * archive,
				   const gchar * dest);
static gchar *archive_extract_tbz2 (const gchar * archive,
				    const gchar * dest);

static ArchiveExtractFunc archive_extract_funcs[] = {
    NULL,
    NULL,
    archive_extract_tar,
    archive_extract_tgz,
    archive_extract_zip,
    archive_extract_tbz2
};


/* FIXME: these functions can be generalised into a function using a
 * command lookup table */

static gchar *archive_extract_tar (const gchar * archive,
				   const gchar * dest)
{
    return g_strdup_printf ("%s >/dev/null xf \"%s\" -C %s", "tar",
			    archive, dest);
}

static gchar *archive_extract_zip (const gchar * archive,
				   const gchar * dest)
{
    return g_strdup_printf ("%s >/dev/null -o -j \"%s\" -d %s", "unzip",
			    archive, dest);
}

static gchar *archive_extract_tgz (const gchar * archive,
				   const gchar * dest)
{
    return g_strdup_printf ("%s >/dev/null xzf \"%s\" -C %s", "tar",
			    archive, dest);
}

static gchar *archive_extract_tbz2 (const gchar * archive,
				    const gchar * dest)
{
    return g_strdup_printf ("bzip2 -dc \"%s\" | %s >/dev/null xf - -C %s",
			    archive, "tar", dest);
}


ArchiveType archive_get_type (const gchar * filename)
{
    gint i = 0;

    if (g_file_test (filename, G_FILE_TEST_IS_DIR))
	return ARCHIVE_DIR;

    while (archive_extensions[i].ext)
	{
	    if (g_str_has_suffix (filename, archive_extensions[i].ext))
		{
		    return archive_extensions[i].type;
		}
	    i++;
	}

    return ARCHIVE_UNKNOWN;
}

gboolean file_is_archive (const gchar * filename)
{
    return (archive_get_type (filename) > ARCHIVE_DIR);
}

gchar *archive_basename (const gchar * str)
{
    gint i = 0;

    while (archive_extensions[i].ext)
	{
	    if (str_has_suffix_nocase (str, archive_extensions[i].ext))
		{
		    const gchar *end = g_strrstr (str, archive_extensions[i].ext);
		    if (end)
			{
			    return g_strndup (str, end - str);
			}
		    break;
		}
	    i++;
	}

    return NULL;
}

/*
   decompress_archive

   Decompresses the archive "filename" to a temporary directory,
   returns the path to the temp dir, or NULL if failed,
   watch out tho, doesn't actually check if the system command succeeds :-|
   */

gchar *archive_decompress (const gchar * filename)
{
    gchar *tmpdir, *cmd, *escaped_filename;
    ArchiveType type;

    if ((type = archive_get_type (filename)) <= ARCHIVE_DIR)
	return NULL;

    tmpdir = g_build_filename (g_get_tmp_dir (), "bmp.XXXXXXXX", NULL);
    if (! mkdtemp (tmpdir))
	{
	    g_free (tmpdir);
	    g_message
	      ("Unable to load skin: Failed to create temporary directory: %s",
	       g_strerror (errno));
	    return NULL;
	}

    escaped_filename = escape_shell_chars (filename);
    cmd = archive_extract_funcs[type] (escaped_filename, tmpdir);
    g_free (escaped_filename);

    if (! cmd)
	{
	    g_message ("extraction function is NULL!");
	    g_free (tmpdir);
	    return NULL;
	}

    system (cmd);
    g_free (cmd);

    return tmpdir;
}


#ifdef HAVE_FTS_H

void del_directory (const gchar * dirname)
{
    gchar *const argv[2] = { (gchar *) dirname, NULL };
    FTS *fts;
    FTSENT *p;

    fts = fts_open (argv, FTS_PHYSICAL, (gint (*) ())NULL);
    while ((p = fts_read (fts)))
	{
	    switch (p->fts_info)
		{
		case FTS_D:
		    break;
		case FTS_DNR:
		case FTS_ERR:
		    break;
		case FTS_DP:
		    rmdir (p->fts_accpath);
		    break;
		default:
		    unlink (p->fts_accpath);
		    break;
		}
	}
    fts_close (fts);
}

#else                           /* !HAVE_FTS */

gboolean del_directory_func (const gchar * path, const gchar * basename,
			     gpointer params)
{
    if (! strcmp (basename, ".") || !strcmp (path, ".."))
	return FALSE;

    if (g_file_test (path, G_FILE_TEST_IS_DIR))
	{
	    dir_foreach (path, del_directory_func, NULL, NULL);
	    rmdir (path);
	    return FALSE;
	}

    unlink (path);

    return FALSE;
}

void del_directory (const gchar * path)
{
    dir_foreach (path, del_directory_func, NULL, NULL);
    rmdir (path);
}

#endif                          /* ifdef HAVE_FTS */

gboolean
is_audio_file (const gchar *filename)
{
    static gchar *audio_files_suffixes[] = {
	"m4a",
	"mp3",
	"mpc",
	"ogg",
	"wma",
	NULL
    };

    if (str_has_suffixes_nocase (filename, audio_files_suffixes)) return TRUE;

    return FALSE;
}

gboolean
collect_path (const gchar * path,
	      const gchar * basename,
	      GList **items)
{
    GError *error = NULL;

    if (g_file_test (path, G_FILE_TEST_IS_REGULAR))
	{
	    if (! is_audio_file (basename)) return FALSE;
	    *items = g_list_append (*items, g_strdup (path));
	}
    else if (g_file_test (path, G_FILE_TEST_IS_DIR))
	{
	    dir_foreach (path, (DirForeachFunc)collect_path,
			 items, &error);
	}

    return FALSE;
}

gchar*
read_ini_string_buffer (const gchar * buffer, const gchar * section,
			const gchar * key)
{
    gchar *ret_buffer = NULL;
    gint found_section = 0, off = 0, len = 0;
    gsize filesize;

    filesize = strlen (buffer);

    if (! buffer)
	return NULL;

    while (! ret_buffer && off < filesize)
	{
	    while (off < filesize &&
		   (buffer[off] == '\r' || buffer[off] == '\n' ||
		    buffer[off] == ' ' || buffer[off] == '\t'))
		off++;
	    if (off >= filesize)
		break;
	    if (buffer[off] == '[')
		{
		    gint slen = strlen (section);
		    off++;
		    found_section = 0;
		    if (off + slen + 1 < filesize &&
			!strncasecmp (section, &buffer[off], slen))
			{
			    off += slen;
			    if (buffer[off] == ']')
				{
				    off++;
				    found_section = 1;
				}
			}
		}
	    else if (found_section && off + strlen (key) < filesize &&
		     !strncasecmp (key, &buffer[off], strlen (key)))
		{
		    off += strlen (key);
		    while (off < filesize &&
			   (buffer[off] == ' ' || buffer[off] == '\t'))
			off++;
		    if (off >= filesize)
			break;
		    if (buffer[off] == '=')
			{
			    off++;
			    while (off < filesize &&
				   (buffer[off] == ' ' || buffer[off] == '\t'))
				off++;
			    if (off >= filesize)
				break;
			    len = 0;
			    while (off + len < filesize &&
				   buffer[off + len] != '\r' &&
				   buffer[off + len] != '\n' &&
				   buffer[off + len] != ';')
				len++;
			    ret_buffer = g_strndup (&buffer[off], len);
			    off += len;
			}
		}
	    while (off < filesize &&
		   buffer[off] != '\r' && buffer[off] != '\n')
		off++;
	}

    return ret_buffer;
}


gchar*
read_ini_string (const gchar * filename, const gchar * section,
		 const gchar * key)
{
    gchar *buffer, *ret_buffer = NULL;
    gint found_section = 0, off = 0, len = 0;
    gsize filesize;

    if (! filename)
	return NULL;

    if (! g_file_get_contents (filename, &buffer, &filesize, NULL))
	return NULL;

    while (! ret_buffer && off < filesize)
	{
	    while (off < filesize &&
		   (buffer[off] == '\r' || buffer[off] == '\n' ||
		    buffer[off] == ' ' || buffer[off] == '\t'))
		off++;
	    if (off >= filesize)
		break;
	    if (buffer[off] == '[')
		{
		    gint slen = strlen (section);
		    off++;
		    found_section = 0;
		    if (off + slen + 1 < filesize &&
			! strncasecmp (section, &buffer[off], slen))
			{
			    off += slen;
			    if (buffer[off] == ']')
				{
				    off++;
				    found_section = 1;
				}
			}
		}
	    else if (found_section && off + strlen (key) < filesize &&
		     ! strncasecmp (key, &buffer[off], strlen (key)))
		{
		    off += strlen (key);
		    while (off < filesize &&
			   (buffer[off] == ' ' || buffer[off] == '\t'))
			off++;
		    if (off >= filesize)
			break;
		    if (buffer[off] == '=')
			{
			    off++;
			    while (off < filesize &&
				   (buffer[off] == ' ' || buffer[off] == '\t'))
				off++;
			    if (off >= filesize)
				break;
			    len = 0;
			    while (off + len < filesize &&
				   buffer[off + len] != '\r' &&
				   buffer[off + len] != '\n' &&
				   buffer[off + len] != ';')
				len++;
			    ret_buffer = g_strndup (&buffer[off], len);
			    off += len;
			}
		}
	    while (off < filesize &&
		   buffer[off] != '\r' && buffer[off] != '\n')
		off++;
	}

    g_free (buffer);
    return ret_buffer;
}

static GArray*
string_to_garray (const gchar * str)
{
    GArray *array;
    gint temp;
    const gchar *ptr = str;
    gchar *endptr;

    array = g_array_new (FALSE, TRUE, sizeof (gint));
    for (;;)
	{
	    temp = strtol (ptr, &endptr, 10);
	    if (ptr == endptr)
		break;
	    g_array_append_val (array, temp);
	    ptr = endptr;
	    while (! isdigit (*ptr) && (*ptr) != '\0')
		ptr++;
	    if (*ptr == '\0')
		break;
	}
    return (array);
}

GArray *read_ini_array (const gchar * filename, const gchar * section,
			const gchar * key)
{
    gchar *temp;
    GArray *a;

    if ((temp = read_ini_string (filename, section, key)) == NULL)
	{
	    g_free (temp);
	    return NULL;
	}
    a = string_to_garray (temp);
    g_free (temp);
    return a;
}

void glist_movedown (GList * list)
{
    gpointer temp;

    if (g_list_next (list))
	{
	    temp = list->data;
	    list->data = list->next->data;
	    list->next->data = temp;
	}
}

void glist_moveup (GList * list)
{
    gpointer temp;

    if (g_list_previous (list))
	{
	    temp = list->data;
	    list->data = list->prev->data;
	    list->prev->data = temp;
	}
}

void
glist_swap (GList * list1, GList * list2)
{
    gpointer aux;

    aux = list1->data;

    list1->data = list2->data;
    list2->data = aux;
}

GList *
glist_make_shuffle_index (GList * list_in)
{
    gint len = g_list_length (list_in);
    gint i, j, n;
    GList *list = NULL, *node, **ptrs;

    if (! len)
	return NULL;

    for (n = 0; n < g_list_length (list_in); n++)
	list = g_list_append (list, GINT_TO_POINTER (n));

    ptrs = g_new (GList *, len);

    for (node = list, i = 0; i < len; node = g_list_next (node), i++)
	ptrs[i] = node;

    j = g_random_int_range (0, len);
    list = ptrs[j];
    ptrs[j]->next = NULL;
    ptrs[j] = ptrs[0];

    for (i = 1; i < len; i++)
	{
	    j = g_random_int_range (0, len - i);
	    list->prev = ptrs[i + j];
	    ptrs[i + j]->next = list;
	    list = ptrs[i + j];
	    ptrs[i + j] = ptrs[i];
	}
    list->prev = NULL;

    g_free (ptrs);

    return list;
}


typedef struct {
    gint x;
    gint y;
} MenuPos;

void
util_menu_position (GtkMenu * menu,
		    gint * x,
		    gint * y,
		    gboolean * push_in,
		    gpointer data)
{
    GtkRequisition requisition;
    gint screen_width;
    gint screen_height;
    MenuPos *pos = data;

    gtk_widget_size_request (GTK_WIDGET (menu), &requisition);

    screen_width = gdk_screen_width ();
    screen_height = gdk_screen_height ();

    *x = CLAMP (pos->x - 2, 0, MAX (0, screen_width - requisition.width));
    *y = CLAMP (pos->y - 2, 0, MAX (0, screen_height - requisition.height));
}

static void util_menu_delete_popup_data (GtkObject * object,
					 GtkItemFactory * ifactory)
{
    gtk_signal_disconnect_by_func (object,
				   GTK_SIGNAL_FUNC
				   (util_menu_delete_popup_data), ifactory);
    gtk_object_remove_data_by_id (GTK_OBJECT (ifactory), quark_popup_data);
}

#ifdef ENABLE_NLS
gchar*
bmp_menu_translate (const gchar * message, gpointer func_data)
{
    gchar *translation = gettext (message);

    if (! translation )
	{
	    g_warning ("Bad translation for message: %s", message);
	    translation = (gchar *) message;
	}

    return translation;
}
#endif

void
util_set_cursor (GtkWidget * window)
{
    static GdkCursor *cursor;

    if (! window)
	{
	    if (cursor)
		{
		    gdk_cursor_destroy (cursor);
		    cursor = NULL;
		}
	    return;
	}
    if (! cursor)
	cursor = gdk_cursor_new (GDK_LEFT_PTR);

    gdk_window_set_cursor (window->window, cursor);
}

/* counts number of digits in a gint */
guint
gint_count_digits (gint n)
{
    guint count = 0;

    n = ABS (n);
    do {
	count++;
	n /= 10;
    } while (n > 0);

    return count;
}

gchar *
str_replace_char (gchar * str, gchar old, gchar new)
{
    gchar *match;

    g_return_val_if_fail (str != NULL, NULL);

    match = str;
    while ((match = strchr (match, old)))
	*match = new;

    return str;
}

gchar *
str_replace (gchar *str,
	     gchar *new_str)
{
    g_free (str);
    return new_str;
}

gboolean
str_has_prefix_nocase (const gchar * str, const gchar * prefix)
{
    return (strncasecmp (str, prefix, strlen (prefix)) == 0);
}

gboolean
str_has_suffix_nocase (const gchar * str, const gchar * suffix)
{
    return (strcasecmp (str + strlen (str) - strlen (suffix), suffix) == 0);
}

gboolean
str_has_suffixes_nocase (const gchar * str,
			 gchar * const *suffixes)
{
    gchar *const *suffix;

    g_return_val_if_fail (str != NULL, FALSE);
    g_return_val_if_fail (suffixes != NULL, FALSE);

    for (suffix = suffixes; *suffix; suffix++)
	if (str_has_suffix_nocase (str, *suffix))
	    return TRUE;

    return FALSE;
}

gboolean
dir_foreach (const gchar * path, DirForeachFunc function,
	     gpointer user_data, GError ** error)
{
    GError *error_out = NULL;
    GDir *dir;
    const gchar *entry;
    gchar *entry_fullpath;

    if (! (dir = g_dir_open (path, 0, &error_out)))
	{
	    g_propagate_error (error, error_out);
	    return FALSE;
	}

    while ((entry = g_dir_read_name (dir)))
	{
	    entry_fullpath = g_build_filename (path, entry, NULL);

	    if ((*function) (entry_fullpath, entry, user_data))
		{
		    g_free (entry_fullpath);
		    break;
		}

	    g_free (entry_fullpath);
	}

    g_dir_close (dir);

    return TRUE;
}

GtkWidget *
ui_manager_get_popup (GtkUIManager * self,
		      const gchar * path)
{
    GtkWidget *menu_item;
    gpointer submenu;

    menu_item = gtk_ui_manager_get_widget (self, path);

    if (GTK_IS_MENU_ITEM (menu_item))
	submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (menu_item));
    else
	submenu = NULL;

    if (submenu)
	{
	    SET_CURSOR_FOR_WIDGET (CURSOR_NORMAL, GTK_WIDGET (submenu));
	}

    return submenu;
}

static void
menu_popup_pos_func (GtkMenu * menu,
		     gint * x,
		     gint * y,
		     gboolean * push_in,
		     gint * point)
{
    *x = point[0];
    *y = point[1];
    *push_in = FALSE;
}

void
bmp_menu_popup (GtkMenu * menu,
		gint x,
		gint y,
		guint button,
		guint time)
{
    gint pos[2];

    pos[0] = x;
    pos[1] = y;

    gtk_menu_popup (menu, NULL, NULL,
		    (GtkMenuPositionFunc) menu_popup_pos_func, pos,
		    button, time);
}

gboolean
bmp_become_daemon (void)
{
    pid_t child_pid;
    int dev_null_fd;

    if (chdir ("/") < 0)
	{
	    g_message ("Could not chdir () to root directory");
	    return FALSE;
	}

    switch ((child_pid = fork ()))
	{
	case -1:
	    g_message ("fork failed\n");
	    return FALSE;
	    break;

	case 0:
	    /* silently ignore failures here, if someone
	     * doesn't have /dev/null we may as well try
	     * to continue anyhow
	     */

	    dev_null_fd = open ("/dev/null", O_RDWR);
	    if (dev_null_fd >= 0)
		{
		    dup2 (dev_null_fd, 0);
		    dup2 (dev_null_fd, 1);
		    dup2 (dev_null_fd, 2);
		}

	    /* Get a predictable umask */
	    umask (022);
	    break;

	default:
	    exit (0);
	    break;
	}

    if (setsid () == -1)
	g_error ("setsid () failed");

    return TRUE;
}

struct MemoryStruct {
    char *memory;
    size_t size;
};

static void *myrealloc (void *ptr, size_t size)
{
    /* There might be a realloc () out there that doesn't like reallocing
       NULL pointers, so we take care of it here */
    if (ptr)
	return realloc (ptr, size);
    else
	return malloc (size);
}

static size_t
WriteMemoryCallback (void *ptr, size_t size, size_t nmemb, void *data)
{
    size_t realsize = size * nmemb;
    struct MemoryStruct *mem = (struct MemoryStruct *)data;

    mem->memory = (char *)myrealloc (mem->memory, mem->size + realsize + 1);
    if (mem->memory)
	{
	    memcpy (& (mem->memory[mem->size]), ptr, realsize);
	    mem->size += realsize;
	    mem->memory[mem->size] = 0;
	}
    return realsize;
}

gchar*
bmp_get_file_http (const gchar *url, GError *error)
{
    CURL *curl_handle;

    struct MemoryStruct chunk;

    chunk.memory = NULL; /* we expect realloc (NULL, size) to work */
    chunk.size = 0;    /* no data at this point */

    curl_global_init (CURL_GLOBAL_ALL);

    /* init the curl session */
    curl_handle = curl_easy_init ();

    /* specify URL to get */
    curl_easy_setopt (curl_handle, CURLOPT_URL, url);

    /* send all data to this function  */
    curl_easy_setopt (curl_handle, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);

    /* we pass our 'chunk' struct to the callback function */
    curl_easy_setopt (curl_handle, CURLOPT_WRITEDATA, (void *)&chunk);

    /* some servers don't like requests that are made without a user-agent
       field, so we provide one */
    curl_easy_setopt (curl_handle, CURLOPT_USERAGENT, "libcurl-agent/1.0");

    /* get it! */
    curl_easy_perform (curl_handle);

    /* cleanup curl stuff */
    curl_easy_cleanup (curl_handle);

    /*
     * Now, our chunk.memory points to a memory block that is chunk.size
     * bytes big and contains the remote file.
     *
     * Do something nice with it!
     *
     * You should be aware of the fact that at this point we might have an
     * allocated data block, and nothing has yet deallocated that data. So when
     * you're done with it, you should free () it as a nice application.
     */

    return chunk.memory;
}

/* Playlist parsing functions */
GList*
get_uris__m3u (const gchar *uri)
{
    GList *uri_list = NULL;
    GError *error = NULL;
    gchar *buffer;
    gchar **uris;
    gint n = 0;
    GURI *g_uri;

    g_return_val_if_fail (uri != NULL, NULL);

    g_uri =
      gnet_uri_new (uri);

    if (! G_URI_IS_LOCAL (uri) && g_uri_test (uri, G_URI_TEST_PROTOCOL, PROTOCOL_HTTP))
	{
	    buffer = bmp_get_file_http (uri, NULL);
	    g_return_val_if_fail (buffer != NULL, NULL);
	}
    else
	{
	    gnet_uri_unescape (g_uri);
	    g_file_get_contents (g_uri->path, &buffer, NULL, &error);
	    if (error)
		{
		    g_message ("%s: ERROR loading '%s': %s", G_STRFUNC, g_uri->path, error->message);
		    return NULL;
		}
	}

    uris =
      g_strsplit (buffer, "\n", 0);
    g_free (buffer);

    while (uris[n])
	{
	    gchar *uri = uris[n];

	    /* Strip leading spaces/tabs */
	    while ((uri[0] == ' ') || (uri[0] == '\t'))
		uri++;
	    /* Skip comments, extended information or null lenght uris */
	    if ((uri[0] == '#') || (uri[0] == '\0'))
		{
		    n++;
		    continue;
		}

	    uri_list = g_list_append (uri_list, g_strdup (uri));
	    n++;
	}

    g_strfreev (uris);
    gnet_uri_delete (g_uri);

    return uri_list;
}


GList*
get_uris__pls (const gchar *uri)
{
    GURI *g_uri;
    GList *uri_list = NULL;
    gchar *num_entries;
    gint num_entries_i, n;

    g_uri =
      gnet_uri_new (uri);

    /* Remote HTTP .pls */
    if (! G_URI_IS_LOCAL (uri) && g_uri_test (uri, G_URI_TEST_PROTOCOL, PROTOCOL_HTTP))
	{
	    gchar *buffer = bmp_get_file_http (uri, NULL);
	    if (! buffer) return NULL;

	    num_entries =
	      read_ini_string_buffer (buffer, "playlist", "numberofentries");
	    num_entries_i =
	      strtol (num_entries, NULL, 10);
	    g_free (num_entries);

	    for (n = 0; n < num_entries_i; n++)
		{
		    gchar *value, *key;

		    key =
		      g_strdup_printf ("File%d", n+1);
		    value =
		      read_ini_string_buffer (buffer, "playlist", key);
		    uri_list =
		      g_list_append (uri_list, g_strdup (value));
		}

	    /* Locally accessible .pls file */
	}
    else
	{
	    num_entries =
	      read_ini_string (g_uri->path, "playlist", "numberofentries");
	    num_entries_i =
	      strtol (num_entries, NULL, 10);
	    g_free (num_entries);

	    for (n = 0; n < num_entries_i; n++)
		{
		    gchar *value, *key;

		    key =
		      g_strdup_printf ("File%d", n+1);
		    value =
		      read_ini_string (g_uri->path, "playlist", key);
		    uri_list =
		      g_list_append (uri_list, g_strdup (value));
		}
	}

    return uri_list;
}

GList*
get_uris__playlist (const gchar *uri)
{
    GList *uri_list = NULL;

    g_return_val_if_fail (uri != NULL, NULL);

    if (str_has_suffix_nocase (uri, ".pls"))
	{
	    uri_list =
	      get_uris__pls (uri);
	}
    else if (str_has_suffix_nocase (uri, ".m3u"))
	{
	    uri_list =
	      get_uris__m3u (uri);
	}

    return uri_list;
}

gboolean
make_directory (const gchar *path,
		mode_t       mode)
{
    if (mkdir (path, mode) == 0)
	return FALSE;

    if (errno == EEXIST)
	return FALSE;

    return TRUE;
}

/* NOTE: This is a shallow copy, it only copies the pointers, not the strings themselves */
gchar**
gslist_to_strv (GSList *list)
{
    gchar **strv;
    gint length, n = 0;

    length = 
      g_slist_length (list)+1;
    strv = 
      g_new (gchar *, length); 

    while (list)
	{
	    strv[n] = list->data;
	    n++;
	    list = list->next;
	}

    strv[n] = NULL;	

    return strv;
}

/* NOTE: This is a shallow copy, it only copies the pointers, not the strings themselves */
gchar**
glist_to_strv (GList *list)
{
    gchar **strv;
    gint length, n = 0;

    length = 
      g_list_length (list)+1;
    strv = 
      g_new (gchar *, length); 

    while (list)
	{
	    strv[n] = list->data;
	    n++;
	    list = list->next;
	}

    strv[n] = NULL;	

    return strv;
}

